#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "ruleadd.h"

#define COLS 13
#define MAX_RULES 100

static struct rule_params *rules[MAX_RULES];/* array containing current rules */
static GtkWidget **table_contents[MAX_RULES];
static int nrules = 0; /* number of rules */

static GtkWidget *table; /* the main table */
void display_table();

/* widgets to contain the table */
static GtkWidget *outer_table;
static GtkWidget *scroll;
static GtkWidget *main_box;
static GtkWidget *window;



/* used internall to execute the external app, with supplied args */
/* should only be used in child process */
void static exe_args(int argc, char **args){
	args[argc] = (void*)0;
	execvp("faultctrl", args); 

	g_print("ERROR: Child process failed\n");
	exit(-1);
}

/* executes command associated with rules */
void execute_command(struct rule_params *r){
	int pid;
	char *args[25]; /* max number of args */
	int argc;
	int i;

	/* sanity check */
	if (!valid_rule(r)){
		display_message("Illegal Rule!");
		return;
	}

	/* fork off */
	pid = fork();

	if (pid > 0){
		/* parent thread, do nothing */
		wait(NULL);
		return;
	}
	else if (pid < 0){
		display_message("Error executing command");
		exit(-1);
	}


	/* allocate some memory for these args */
	for (i=0; i<25; i++) args[i] = malloc(sizeof(char)*10);

	sprintf(args[0], "faultctrl");

	/* we are in child process */
	if (r->operation == FLUSH){
		sprintf(args[1], "-F");
		exe_args(2, args); /* we're done for the flush */
		exit(0);
	}

	if (r->operation == DELETE){
		sprintf(args[1], "-D");
		sprintf(args[2], "%d", r->rule_num);
		exe_args(3, args);
		exit(0);
	}

	if (r->operation == APPEND) sprintf(args[1], "-A");
	else if (r->operation == INSERT) sprintf(args[1], "-I");
	else if (r->operation == UPDATE) sprintf(args[1], "-U");
	argc = 2; /* one arg so far */

	if (r->operation == INSERT || r->operation == UPDATE){
		/* specify the rule num to update or insert */
		sprintf(args[argc], "%d", r->rule_num);
		argc++;
	}

	/* specify the rule type */
	if (r->rule_type == DROP){
		sprintf(args[argc], "drop");
		argc++;
	}
	else if (r->rule_type == REORDER){
		sprintf(args[argc], "reorder");
		argc++;
	}
	else{
		g_print("Error: unknown type");
		exit(0);
	}

	/* specify protocol (this is optional argument, but its given by GUI */
	if (r->proto == PROTO_IGMP){
		sprintf(args[argc], "-p");
		argc++;
		sprintf(args[argc], "igmp");
		argc++;
	}
	else if (r->proto == PROTO_ICMP){
		sprintf(args[argc], "-p");
		argc++;
		sprintf(args[argc], "icmp");
		argc++;
	}
	else if (r->proto == PROTO_TCP){
		sprintf(args[argc], "-p");
		argc++;
		sprintf(args[argc], "tcp");
		argc++;
	}
	else if (r->proto == PROTO_UDP){
		sprintf(args[argc], "-p");
		argc++;
		sprintf(args[argc], "udp");
		argc++;
	}
	/* else it is for all protos -> ignore this arg */

	sprintf(args[argc], "-min");
	argc++;
	sprintf(args[argc], "%d", r->min_size);
	argc++;
	sprintf(args[argc], "-max");
	argc++;
	sprintf(args[argc], "%d", r->max_size);
	argc++;

	/* if reordering, specify burst and delay params */
	if (r->rule_type == REORDER){
		sprintf(args[argc], "-maxburst");
		argc++;
		sprintf(args[argc], "%d", r->max_burst_size);
		argc++;
		sprintf(args[argc], "-maxdelay");
		argc++;
		sprintf(args[argc], "%d", r->max_delay_size);
		argc++;

		sprintf(args[argc], "-burstskew");
		argc++;
		sprintf(args[argc], "%d", r->burst_skew);
		argc++;
		sprintf(args[argc], "-delayskew");
		argc++;
		sprintf(args[argc], "%d", r->delay_skew);
		argc++;
	}

	/* record probability */
	sprintf(args[argc], "%d", r->probability);
	argc++;

	exe_args(argc, args);
	exit(0); /* leave child process */
}

/* checks the rule for validity */
int valid_rule(struct rule_params *r){
	if (r == NULL) return 0;
	if (r->operation == FLUSH) return 1; /* legal already */

	if (r->operation < APPEND || r->operation > FLUSH) return 0;
	if (r->operation == INSERT ||
		r->operation == UPDATE ||
		r->operation == DELETE){
			if (r->rule_num > nrules || r->rule_num < 0) return 0;
	}

	if (r->operation == DELETE){ 
		if (r->rule_num < 0 || r->rule_num > nrules) return 0;
		/* now legal for DELETE operations */
		return 1;
	}

	if (!
		(r->proto == PROTO_ALL || r->proto == PROTO_ICMP ||
		 r->proto == PROTO_IGMP || r->proto == PROTO_TCP ||
		 r->proto == PROTO_UDP)
		) return 0;
	if (r->operation != APPEND)
		if (r->rule_num < 0 || r->rule_num >= MAX_RULES) return 0;
	if (!(r->rule_type == DROP || r->rule_type == REORDER)) return 0;
	if (r->min_size < 0) return 0;
	if (r->max_size <= r->min_size) return 0;
	if (r->rule_type == REORDER){
		if (r->max_burst_size < 1) return 0;
		if (r->burst_skew < -100 || r->burst_skew > 100) return 0;
		if (r->max_delay_size < 1) return 0;
		if (r->delay_skew < -100 || r->delay_skew > 100) return 0;
	}
	if (r->probability < 1 || r->probability > 1000) return 0;

	return 1; /* all checks passed */
}


void copy_rule(struct rule_params *to, struct rule_params *from){
	to->rule_type = from->rule_type;
	to->operation = from->operation;
	to->proto = from->proto;
	to->rule_num = from->rule_num;
	to->min_size = from->min_size;
	to->max_size = from->max_size;
	to->max_burst_size = from->max_burst_size;
	to->burst_skew = from->burst_skew;
	to->max_delay_size = from->max_delay_size;
	to->delay_skew = from->delay_skew;
	to->probability = from->probability;
}


/* appends rule to end of list */
void append_rule(struct rule_params *r){

	/* sanity checks */
	if (nrules >= MAX_RULES){
		display_message("Maximum Number of Rules Reached");
		return;
	}
	if (!valid_rule(r)){
		display_message("Error: Invalid rule");
		return;
	}

	rules[nrules] = malloc(sizeof(struct rule_params));
	copy_rule(rules[nrules], r);

	nrules++;

	execute_command(r);
	display_table(); /* redisplay the table */
}

/* insert a rule into the list of rules */
void insert_rule(struct rule_params *r){
	int pos, end;
	
	/* sanity checks */
	if (nrules >= MAX_RULES){
		display_message("Maximum Number of Rules Reached");
		return;
	}
	if (!valid_rule(r)){
		display_message("Error: Invalid rule");
		return;
	}

	pos = r->rule_num; /* position to insert at */
	end = nrules;

	while(end > pos){
		rules[end] = rules[end-1];
		end--;
	}

	rules[pos] = malloc(sizeof(struct rule_params));
	copy_rule(rules[pos], r);

	nrules++;

	execute_command(r);
	display_table(); /* redisplay the table */
}


/* updates a current rule, with a ne one. overwritting values */
void update_rule(struct rule_params *r){
	/* sanity */
	if (!valid_rule(r)){
		display_message("Error in update");
		return;
	}

	/* simply overwrite the rule */
	copy_rule(rules[r->rule_num], r);

	execute_command(r);
	display_table(); /* redisplay the table */
}


/* delete a rule from the list */
void delete_rule(struct rule_params *r){
	int pos;

	/* sanity check */
	if (!valid_rule(r)){
		display_message("Cannot delete");
		return;
	}

	pos = r->rule_num;
	free(rules[pos]);

	while (pos < nrules-1){
		rules[pos] = rules[pos+1];
		pos++;
	}

	nrules--;

	execute_command(r);
	display_table(); /* redisplay the table */
}


void flush_rules(struct rule_params *r){
	/* free up our allocated memory */
	int i;
	for (i=0; i<nrules; i++){
		free(rules[i]);
	}

	nrules = 0; /* reset count */

	execute_command(r);
	display_table(); /* redisplay the table */
}


void rule_in(struct rule_params *r){
	if (r->operation == APPEND) append_rule(r);
	else if (r->operation == INSERT) insert_rule(r);
	else if (r->operation == UPDATE) update_rule(r);
	else if (r->operation == DELETE) delete_rule(r);
	else if (r->operation == FLUSH) flush_rules(r);
	else{
		g_print("Unknown Rule Operation\n");
	}
}



/***
* SIGNAL HANDLERS
***/


/*
*	Main window Buttons
*/
static void add_button_pressed(GtkWidget *widget, gpointer data){
	struct rule_params *r = malloc(sizeof(struct rule_params));
	r->operation = APPEND;
	addrule_window(r);
}

static void clear_button_pressed(GtkWidget *widget, gpointer data){
	struct rule_params *r = malloc(sizeof(struct rule_params));
	r->operation = FLUSH;
	rule_in(r);
	free(r);
}

static void exit_button_pressed(GtkWidget *widget, gpointer data){
	struct rule_params *r = malloc(sizeof(struct rule_params));
	r->operation = FLUSH;

	/* clear all rules */
	flush_rules(r);

	/* get all out */
	gtk_main_quit();
}

/* 
*		Buttons associated with each rule 
*/
static void update_button_pressed(GtkWidget *widget, gpointer data){
	struct rule_params *r = malloc(sizeof(struct rule_params));

	/* sanity check */
	if ((nrules <= (int)data) || (0 > (int)data)){
		display_message("Error: Rule is NULL");
		return;
	}

	copy_rule(r, rules[(int)data]);
	r->operation = UPDATE;
	r->rule_num = (int)data;
	addrule_window(r);
}

static void insert_button_pressed(GtkWidget *widget, gpointer data){
	struct rule_params *r = malloc(sizeof(struct rule_params));
	r->operation = INSERT;
	r->rule_num = (int)data;
	addrule_window(r);
}

static void delete_button_pressed(GtkWidget *widget, gpointer data){
	struct rule_params *r = malloc(sizeof(struct rule_params));
	r->operation = DELETE;
	r->rule_num = (int)data;

	/* clear all rules */
	rule_in(r);
	free(r);
}





/*******
*		GUI components
********/

/* displays one row of the table
*params: GtkTable, rule number to display */
void display_table_row(GtkWidget *t, int data_row){
	char tmp[20];
	GtkTooltips *tooltip;
	int col = 0;

	/* sanity checks */
	if (t == NULL){
		g_print("Error [display_table_row]:\tTable is NULL\n");
		return;
	}
	if (data_row < 0 || data_row >= nrules){
		g_print
			("Error [display_table_row]:\tCannot display row: %d\n", data_row);
		return;
	}

	if (table_contents[data_row] == NULL){
		int i;
		table_contents[data_row] = malloc(sizeof(GtkWidget)*COLS);

		for (i=0; i<COLS; i++) table_contents[data_row][i] = NULL;
	}

	/* clean up any old widgets being displayed in this row */
	/*
	while(col < COLS){
		if (table_contents[data_row][col] != NULL){
			gtk_widget_destroy(table_contents[data_row][col]);
			*/

			/* check its deleted */
			/*
			if (table_contents[data_row][col] != NULL){
				g_print("Error [display_table_row:\tWidget not null\n");
			}
		}
		col++;
	}
	*/

	/* insert the widgets into this row, in table_contents */
	table_contents[data_row][0] = gtk_button_new_with_label("U");
	tooltip = gtk_tooltips_new();
	gtk_tooltips_set_tip(tooltip, table_contents[data_row][0], 
						"Update Rule", "");
	gtk_widget_set_usize(table_contents[data_row][0], 30, 30);
	gtk_signal_connect(GTK_OBJECT(table_contents[data_row][0]), "clicked", 
				GTK_SIGNAL_FUNC(update_button_pressed), (gpointer)data_row);

	table_contents[data_row][1] = gtk_button_new_with_label("I");
	tooltip = gtk_tooltips_new();
	gtk_tooltips_set_tip(tooltip, table_contents[data_row][1], 
						"Insert Rule", "");
	gtk_widget_set_usize(table_contents[data_row][1], 30, 30);
	gtk_signal_connect(GTK_OBJECT(table_contents[data_row][1]), "clicked", 
				GTK_SIGNAL_FUNC(insert_button_pressed), (gpointer)data_row);

	table_contents[data_row][2] = gtk_button_new_with_label("X");
	tooltip = gtk_tooltips_new();
	gtk_tooltips_set_tip(tooltip, table_contents[data_row][2], 
						"Delete Rule", "");
	gtk_widget_set_usize(table_contents[data_row][2], 30, 30);
	gtk_signal_connect(GTK_OBJECT(table_contents[data_row][2]), "clicked", 
				GTK_SIGNAL_FUNC(delete_button_pressed), (gpointer)data_row);


	sprintf(tmp, "%d", data_row); /* use data_row as rule_num */
	table_contents[data_row][3] = gtk_entry_new_with_max_length(4);
	gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][3]), tmp);
	gtk_widget_set_usize(table_contents[data_row][3], 40, 25);

	/* set rule type entry */
	table_contents[data_row][4] = gtk_entry_new_with_max_length(8);
	if (rules[data_row]->rule_type == DROP){
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][4]), "DROP");
	}
	else{
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][4]), "REORDER");
	}
	gtk_widget_set_usize(table_contents[data_row][4], 100, 25);

	/* setup widgits for remainder of row */
	col = 5;
	while (col < COLS){
		table_contents[data_row][col] = gtk_entry_new();
		gtk_entry_set_max_length(GTK_ENTRY(table_contents[data_row][col]), 5);
		gtk_widget_set_usize(table_contents[data_row][col], 90, 25);
		col++;
	}

	/* set the protocol entry */
	if (rules[data_row]->proto == PROTO_TCP){
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][5]), "TCP");
	}
	else if (rules[data_row]->proto == PROTO_UDP){
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][5]), "UDP");
	}
	else if (rules[data_row]->proto == PROTO_ICMP){
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][5]), "ICMP");
	}
	else if (rules[data_row]->proto == PROTO_IGMP){
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][5]), "IGMP");
	}
	else{
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][5]), "ALL");
	}

	/* set size info */
	sprintf(tmp, "%d", rules[data_row]->min_size);
	gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][6]), tmp);
	sprintf(tmp, "%d", rules[data_row]->max_size);
	gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][7]), tmp);

	/* set probability */
	sprintf(tmp, "%2.1f%%", ((float)rules[data_row]->probability)/10);
	gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][8]), tmp);


	/* if we are reordering, display burst and delay settings */
	if (rules[data_row]->rule_type == REORDER){
		sprintf(tmp, "%d", rules[data_row]->max_burst_size);
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][9]), tmp);

		sprintf(tmp, "%d%%", rules[data_row]->burst_skew);
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][10]), tmp);


		sprintf(tmp, "%d", rules[data_row]->max_delay_size);
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][11]), tmp);

		sprintf(tmp, "%d%%", rules[data_row]->delay_skew);
		gtk_entry_set_text(GTK_ENTRY(table_contents[data_row][12]), tmp);
	}


	/* set all the entries uneditable */
	col = 3;
	while(col < COLS){
		gtk_entry_set_editable(GTK_ENTRY(table_contents[data_row][col]), 0);
		col++;
	}


	/* insert entries in table */
	col = 0;
	while(col < COLS){
		gtk_table_attach_defaults(
			GTK_TABLE(t),
			table_contents[data_row][col],
			col,
			col+1,
			data_row+1,
			data_row+2
			);

		col++;
	}
}

/* insert column headers into table */
/* parameter is the table */
void display_table_headers(GtkWidget *t){
	GtkWidget *label;
	int i;

	/* padding for buttons */
	for (i = 0; i<3; i++){
		label = gtk_label_new("");
		gtk_widget_set_usize(label, 30, 30);
		gtk_table_attach_defaults(GTK_TABLE(t), label, i, i+1, 0, 1);
	}

	label = gtk_label_new("No");
	gtk_widget_set_usize(label, 40, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 3, 4, 0, 1);

	label = gtk_label_new("Type");
	gtk_widget_set_usize(label, 100, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 4, 5, 0, 1);

	label = gtk_label_new("Proto");
	gtk_widget_set_usize(label, 90, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 5, 6, 0, 1);

	label = gtk_label_new("Min Size");
	gtk_widget_set_usize(label, 90, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 6, 7, 0, 1);

	label = gtk_label_new("Max Size");
	gtk_widget_set_usize(label, 90, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 7, 8, 0, 1);

	label = gtk_label_new("Probability");
	gtk_widget_set_usize(label, 90, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 8, 9, 0, 1);

	label = gtk_label_new("Max Burst");
	gtk_widget_set_usize(label, 90, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 9, 10, 0, 1);

	label = gtk_label_new("Burst Skew");
	gtk_widget_set_usize(label, 90, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 10, 11, 0, 1);

	label = gtk_label_new("Max Delay");
	gtk_widget_set_usize(label, 90, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 11, 12, 0, 1);

	label = gtk_label_new("Delay Skew");
	gtk_widget_set_usize(label, 90, 25);
	gtk_table_attach_defaults(GTK_TABLE(t), label, 12, 13, 0, 1);
}


/* insert window buttons, add/clear/exit */
void display_window_buttons(GtkWidget *box){
	GtkWidget *add_button, *clear_button, *exit_button;

	gtk_box_set_spacing(GTK_BOX(box), 150);

	add_button = gtk_button_new_with_label("Add");
	clear_button = gtk_button_new_with_label("Clear");
	exit_button = gtk_button_new_with_label("Exit");

	gtk_signal_connect(GTK_OBJECT(add_button), "clicked", 
						GTK_SIGNAL_FUNC(add_button_pressed), NULL);
	gtk_signal_connect(GTK_OBJECT(clear_button), "clicked", 
						GTK_SIGNAL_FUNC(clear_button_pressed), NULL);
	gtk_signal_connect(GTK_OBJECT(exit_button), "clicked", 
						GTK_SIGNAL_FUNC(exit_button_pressed), NULL);

	/* insert into box */
	gtk_box_pack_start(GTK_BOX(box), add_button, 0, 0, 40);
	gtk_box_pack_start(GTK_BOX(box), clear_button, 0, 0, 40);
	gtk_box_pack_start(GTK_BOX(box), exit_button, 0, 0, 40);
}


/* handles displaying the table */
void display_table(){
	int i;

	gtk_widget_destroy(table);

	table = gtk_table_new(MAX_RULES, COLS, 0);

	gtk_table_attach_defaults(GTK_TABLE(outer_table), table, 0, 1, 0, 1);
	display_table_headers(table);

	for (i=0; i<nrules; i++){
		display_table_row(table, i);
	}
	gtk_widget_show_all(table);
}


int main(int argc, char **argv){
	GtkWidget *button_box;
	GtkWidget *align;

	gtk_init(&argc, &argv);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "Faulty Network Router");
	gtk_window_set_default_size(GTK_WINDOW(window), 300, 400);

	/* make closing window as an exit */
	gtk_signal_connect(GTK_OBJECT(window), "destroy", 
		               GTK_SIGNAL_FUNC(exit_button_pressed), NULL);

	main_box = gtk_vbox_new(0, 50);
	/* main_box = gtk_table_new(2, 1, 0); */
	button_box = gtk_hbox_new(1, 150);
	table = gtk_table_new(MAX_RULES, COLS, 0);
	outer_table = gtk_table_new(1, 1, 0);
	align = gtk_alignment_new(0.5, 0.0, 0.0, 0.0);
	gtk_container_add(GTK_CONTAINER(align), outer_table);

	gtk_table_attach_defaults(GTK_TABLE(outer_table), table, 0, 1, 0, 1);

	gtk_container_set_resize_mode(GTK_CONTAINER(button_box), GTK_RESIZE_PARENT);

	/* set up the scrollable window */
	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), 
								   GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_add_with_viewport(
									GTK_SCROLLED_WINDOW(scroll), align);


	display_table_headers(table);
	display_table();
	display_window_buttons(button_box);

	gtk_container_add(GTK_CONTAINER(window), main_box);

	gtk_box_pack_start(GTK_BOX(main_box), scroll, 1, 1, 0);
	gtk_box_pack_start(GTK_BOX(main_box), button_box, 0, 0, 0);
	/*
	gtk_table_attach_defaults(GTK_TABLE(main_box), scroll, 0, 1, 0, 1);
	gtk_table_attach_defaults(GTK_TABLE(main_box), button_box, 0, 1, 1, 2);
	*/


	gtk_widget_show_all(window);


	gtk_main();
	return 0;
}
